



if not exist %PREFIX%\\Library\\include\\ignition\\gazebo5\\ignition\\gazebo.hh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\lib\\ignition-gazebo5.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\bin\\ignition-gazebo5.dll exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\lib\\cmake\\ignition-gazebo5\\ignition-gazebo5-config.cmake exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
